from __future__ import annotations

import json
import logging
import os
from typing import Dict, Type, List, Tuple, Callable, Any

import couchdb
import pydash
from couchdb import ResourceConflict, ServerError
from pydantic import BaseModel, Field
from pydantic_db_backend_common.exceptions import (
    NotFound,
    AlreadyExists,
    RevisionConflict,
)
from pydantic_db_backend_common.indexes import Index, SortingIndex, AggregationIndex
from pydantic_db_backend_common.pydantic import BackendModel
from pydantic_settings import BaseSettings

from pydantic_db_backend.backend import Backend
from pydantic_db_backend.utils import CustomJSONEncoder

log = logging.getLogger(__name__)


class CouchDbBackendSettings(BaseSettings):
    couchdb_uri: str


class CouchDbBackend(Backend):
    settings_class = CouchDbBackendSettings
    connection: CouchDbConnectionModel
    # unsupported_features = ["max_results", "distinct"]
    unsupported_features = ["distinct"]

    def __init__(self, alias: str = "default"):
        super().__init__(alias)
        uri = self.settings.couchdb_uri
        self.connection = CouchDbConnectionModel(alias=alias, uri=uri, server=couchdb.Server(uri))

    @classmethod
    def to_db(cls, instance: BackendModel, json_dict: bool | None = True) -> dict:
        document = super().to_db(instance)
        document = pydash.omit(
            document | {"_id": document["uid"], "_rev": document["revision"]},
            "uid",
            "revision",
        )
        return document

    @classmethod
    def from_db(
        cls, model: Type[BackendModel], document: dict, json_dict: bool | None = True
    ) -> BackendModel:
        document = pydash.omit(
            document | {"uid": document["_id"], "revision": document["_rev"]},
            "_id",
            "_rev",
        )
        return super().from_db(model, document)

    def get_collection(self, model: Type[BackendModel]) -> couchdb.Database:
        collection_name = self.collection_name(model)
        con = self.connection

        if collection_name in con.server:
            col = con.server[collection_name]
        else:
            col = con.server.create(collection_name)
            self.indexes(model, dict(collection=col))
        return col

    def delete_collection(self, model: Type[BackendModel]) -> None:
        server = self.connection.server
        name = self.collection_name(model)
        if name in server:
            server.delete(name)
        super().delete_collection(model)

    def post_document(self, model: Type[BackendModel], document: dict) -> Dict:
        col = self.get_collection(model)
        try:
            col.save(pydash.omit(document, "_rev"))
        except ResourceConflict as e:
            raise AlreadyExists(uid=document["_id"])
        return col.get(document["_id"])

    def post_instance(self, instance: BackendModel) -> BackendModel:
        document = self.to_db(instance)
        document = self.post_document(instance.__class__, document)
        return self.from_db(instance.__class__, document)

    def get_instance(self, model: Type[BackendModel], uid: str) -> BackendModel:
        col = self.get_collection(model)
        entry = col.get(uid)
        if entry is None:
            raise NotFound(uid)
        return self.from_db(model, entry)

    def put_instance(self, instance: BackendModel, ignore_revision_conflict: bool = False) -> BackendModel:
        col = self.get_collection(instance.__class__)
        if instance.uid in col:
            document = self.to_db(instance)
            while True:
                try:
                    id, rev = col.save(document)
                    document["_rev"] = rev
                    return self.from_db(instance.__class__, document)

                except ResourceConflict as e:
                    new_rev = col.get(instance.uid)["_rev"]
                    if ignore_revision_conflict:
                        document["_rev"] = new_rev
                        continue
                    raise RevisionConflict(new_rev)

        else:
            return self.post_instance(instance)

    def delete_uid(cls, model: Type[BackendModel], uid: str) -> None:
        col = cls.get_collection(model)
        if uid in col:
            del col[uid]
        else:
            raise NotFound(uid=uid)

    def find(
        self,
        model: Type[BackendModel],
        skip: int = 0,
        limit: int = 25,
        query_filter: dict = None,
        sort: List = None,
        fields: List[str] = None,
        max_results: bool | None = False,
        func: Callable = None,
    ) -> Tuple[List[Any], int] | List[Any]:
        # fix 0 limit, since couchdb does not know this
        limit = 9999999 if limit == 0 else limit

        if query_filter is None:
            query_filter = {}

        # convert to json and back again, to have iso datetime strings
        query_filter = json.loads(json.dumps(query_filter, cls=CustomJSONEncoder))

        col = self.get_collection(model)

        find_dict = {
            "selector": query_filter,
            "skip": skip,
            "limit": limit,
        }

        if fields is not None:
            find_dict |= {"fields": fields}

        if sort is not None:
            find_dict["sort"] = sort

        try:
            find_result = col.find(find_dict)
        except ServerError as e:
            error_code = pydash.get(e, ["args", 0, 0])

            if error_code == 400:
                # @asc:  not what I expected the system to do. Better would be to modify the
                # index cache and initiate a new get_db... and a loop
                self.indexes(model, dict(collection=col), force_index_creation=True)
                find_result = col.find(find_dict)
            else:
                raise e
        result = [func(x) for x in find_result]
        if max_results:
            max_results = self.get_max_results(model, query_filter)
            return result, max_results
        else:
            return result

    def get_max_results(self, model: Type[BackendModel], query_filter: dict) -> int:
        # atm it only works with one field,
        col = self.get_collection(model)

        field = "_".join(list(query_filter.keys()))
        view_name = f"{field}_sum"
        index = self.get_index_by_name(model, view_name)

        design_document = f"aggregation_{view_name}/{view_name}"
        for row in col.view(design_document, group=True):
            if field == "":
                return row.value
            else:
                for k, v in query_filter.items():
                    if isinstance(v, dict):
                        k = next(iter(v.keys()))
                        if k == "$eq":
                            v = v[k]
                        else:
                            return 0
                    if v == row.key:
                        ret = row.value
                        return ret

    # noinspection PyMethodOverriding
    def create_index(self, collection_name: str, index: Index, collection: couchdb.Database):
        super().create_index(collection_name, index)

        if index.type == "sorting":
            index: SortingIndex

            index_key = (collection_name, index.name)
            indexes = collection.index()

            # noinspection PyUnboundLocalVariable
            if index_key not in indexes:
                indexes[index_key] = index.sorting

        elif index.type == "aggregation":
            index: AggregationIndex
            self.view_from_aggregation_index(collection, index)

        else:
            pass

    def view_from_aggregation_index(self, collection: couchdb.Database, index: AggregationIndex):
        design_document = index.design_document

        in_db = f"_design/{design_document}" in collection
        if not in_db:
            field, func = next(iter(index.spec.items()))
            if field is None or field == "":
                map_function = f"function (doc) {{ emit(null, 1); }}"
            else:
                map_function = f"function (doc) {{ emit(doc.{field}, 1); }}"
            view_name = f"{index.view_name}"
            # reduce_function = f"function(keys, values, rereduce) {{ return sum(values); }}"
            data = {
                "_id": f"_design/{design_document}",
                "views": {view_name: {"map": map_function, "reduce": func}},
                "language": "javascript",
                "options": {"partitioned": False},
            }
            logging.info(f"creating view {design_document}/{view_name}")
            collection.save(data)


class CouchDbConnectionModel(BaseModel):
    alias: str
    uri: str
    server: couchdb.Server
    dbs: Dict[str, couchdb.Database] = Field(default_factory=dict)

    class Config:
        arbitrary_types_allowed = True



from __future__ import annotations

import json
import logging
from typing import Dict, Type, List, Tuple, Callable, Any
from urllib.parse import urlparse

import couchdb
import pydash
from bson import SON
from pydantic_db_backend_common.exceptions import (
    NotFound,
    AlreadyExists,
    RevisionConflict,
)
from pydantic_db_backend_common.indexes import AggregationIndex
from pydantic_db_backend_common.pydantic import BackendModel
from pydantic_settings import BaseSettings
from pymongo import MongoClient, collection
from pymongo.errors import DuplicateKeyError

from pydantic_db_backend.backend import Backend
from pydantic_db_backend.utils import CustomJSONEncoder, CustomJSONDecoder

log = logging.getLogger(__name__)


class MongoDbBackendSettings(BaseSettings):
    mongodb_uri: str


class MongoDbBackend(Backend):
    settings_class = MongoDbBackendSettings
    db_name: str
    connection: MongoClient

    def __init__(self, alias: str = "default"):
        super().__init__(alias)
        self.db_name = urlparse(self.settings.mongodb_uri).path.strip("/")
        self.connection = MongoClient(self.settings.mongodb_uri, tz_aware=True)
        self.db = self.connection[self.db_name]

    @classmethod
    def to_db(cls, instance: BackendModel, json_dict: bool | None = False) -> dict:
        document = super().to_db(instance, json_dict=json_dict)
        document = pydash.omit(
            document | {"_id": document["uid"], "_rev": document["revision"]},
            "uid",
            "revision",
        )

        return document

    @classmethod
    def from_db(
        cls, model: Type[BackendModel], document: dict, json_dict: bool | None = False
    ) -> BackendModel:
        document = pydash.omit(
            document | {"uid": document["_id"], "revision": document["_rev"]},
            "_id",
            "_rev",
        )
        return super().from_db(model, document, json_dict)

    def get_collection(self, model: Type[BackendModel]) -> collection:
        collection_name = self.collection_name(model)
        # cls.indexes(model, dict(db=db))
        return self.db[collection_name]

    def delete_collection(self, model: Type[BackendModel]) -> None:
        col = self.get_collection(model)
        col.drop()
        super().delete_collection(model)

    def post_document(self, model: Type[BackendModel], document: dict) -> Dict:
        col = self.get_collection(model)
        try:
            self.set_document_revision(document, field="_rev")
            col.insert_one(document)
        except DuplicateKeyError as e:
            raise AlreadyExists(uid=document["_id"])
        r = col.find_one({"_id": document["_id"]})
        return r

    def post_instance(self, instance: BackendModel) -> BackendModel:
        document = self.to_db(instance)
        document = self.post_document(instance.__class__, document)
        return self.from_db(instance.__class__, document)

    def get_instance(self, model: Type[BackendModel], uid: str) -> BackendModel:
        col = self.get_collection(model)
        entry = col.find_one({"_id": uid})
        if entry is None:
            raise NotFound(uid)
        return self.from_db(model, entry)

    def put_instance(self, instance: BackendModel, ignore_revision_conflict: bool = False) -> BackendModel:
        # get old revision from doc
        # set new revision in doc

        # update doc with old revision

        # if failed, then there are 2 reasons
        #   a) id does not exist anymore.
        #   b) version does not match.

        col = self.get_collection(instance.__class__)
        document = self.to_db(instance)

        old_rev = self.get_document_revision(document)
        self.set_document_revision(document)  # generate new revision hash
        match = {
            "_id": document["_id"],
        }

        if not ignore_revision_conflict:
            match |= {
                "_rev": old_rev,
            }

        r = col.update_one(
            match,
            {"$set": document},
        )
        # did update succeed ?
        if r.modified_count == 1:
            # everything worked.
            return self.from_db(instance.__class__, document)

        else:
            # update didn't work. try to find document in db.
            r = col.find_one({"_id": document["_id"]})
            if r is not None:
                # document exists , check version
                raise RevisionConflict(r["_rev"])
            else:
                # No document, post
                return self.post_instance(instance)

    def delete_uid(self, model: Type[BackendModel], uid: str) -> None:
        col = self.get_collection(model)
        r = col.delete_one({"_id": uid})  # assuming that there is only one
        if r.deleted_count == 0:
            raise NotFound(uid=uid)

    @staticmethod
    def _convert_query_filter(filter: dict) -> dict:
        # {'worker_expires': {'$and': [{'$ne': None}, {'$lt': '2022-01-01T00:10:00+00:00'}]}}
        def convert(d: Any) -> Any:
            if isinstance(d, dict):
                if len(d) == 1 and list(d.keys())[0] == "$and":
                    return pydash.assign({}, *d["$and"])
            return d

        ret = {k: convert(v) for k, v in filter.items()}
        return ret

    def find(
        self,
        model: Type[BackendModel],
        skip: int = 0,
        limit: int = 25,
        query_filter: dict = None,
        sort: List = None,
        fields: List[str] = None,
        max_results: bool | None = False,
        func: Callable = None,
    ) -> Tuple[List[Any], int] | List[Any]:
        # fix 0 limit, since couchdb does not know this
        limit = 9999999 if limit == 0 else limit

        if query_filter is None:
            query_filter = {}

        # convert to json and back again, to have real datetime objects
        query_filter = json.loads(json.dumps(query_filter, cls=CustomJSONEncoder), cls=CustomJSONDecoder)
        query_filter = self._convert_query_filter(query_filter)

        col = self.get_collection(model)

        data = []
        if sort is not None and len(sort) != 0:
            sort = SON([(f, 1 if d.lower() == "asc" else -1) for x in sort for f, d in x.items()])
            data.append({"$sort": sort})

        data.append({"$skip": skip})
        data.append({"$limit": limit})

        # if fields is not None:
        #     find_dict |= {"fields": fields}

        agg = [
            {"$match": query_filter},
            {
                "$facet": {
                    "meta": [{"$group": {"_id": None, "max_results": {"$sum": 1}}}],
                    "data": data,
                }
            },
        ]



        try:
            agg_result = next(iter(col.aggregate(agg)), None)

        except Exception as e:
            raise e

        if agg_result is None:
            result = []
            ret_max_results = 0
        else:
            ret_max_results = pydash.get(agg_result, ["meta", 0, "max_results"], 0)
            result = [func(x) for x in pydash.get(agg_result, ["data"], [])]

        if max_results:
            return result, ret_max_results
        else:
            return result



from __future__ import annotations

import hashlib
import json
import logging
import re
from abc import ABC, abstractmethod
from typing import Type, Dict, List, Tuple, Callable, Any

import pydash
from pydantic_db_backend_common.exceptions import IndexNotExisting
from pydantic_db_backend_common.indexes import Index
from pydantic_db_backend_common.pydantic import BackendModel
from pydantic_db_backend_common.utils import uid, utcnow
from pydantic_settings import BaseSettings

from pydantic_db_backend.utils import CustomJSONEncoder

log = logging.getLogger(__name__)


class Backend(ABC):
    _clients: Dict[str, Backend] = {}
    _collections: Dict[Type[BackendModel], str] = {}
    _revision_ignored_fields: List[str] = ["updated_time"]
    _indexes: Dict[str, list] = {}

    settings_class: Type[BaseSettings] | None = None
    unsupported_features: List[str] = []

    def __init__(self, alias: str = "default"):
        self.settings = self.settings_class() if self.settings_class is not None else {}
        # noinspection PyTypeChecker
        Backend.set_client(alias, self)

    @classmethod
    def set_client(cls, alias: str, backend: Backend):
        cls._clients[alias] = backend

    @classmethod
    def client(cls, alias: str = "default") -> Backend:
        # todo: raise error if alias not known
        return cls._clients[alias]

    @classmethod
    def collection_name(cls, model: Type[BackendModel]) -> str:
        if model not in cls._collections:
            name = (
                re.sub("([A-Z]+)", r"_\1", model.__name__).lower().removeprefix("_").removesuffix("_model")
            )
            cls._collections[model] = name
        return cls._collections[model]

    @classmethod
    def get_document_revision(cls, document: dict, field: str = "_rev") -> str:
        return pydash.default_to(pydash.get(document, field, None), "0-0")

    @classmethod
    def set_document_revision(cls, document: dict, field: str = "_rev"):
        old_rev, old_checksum = cls.get_document_revision(document).split("-")
        document_data = pydash.omit(document, field, *cls._revision_ignored_fields)
        checksum = hashlib.md5(
            json.dumps(document_data, cls=CustomJSONEncoder, sort_keys=True).encode("utf8")
        ).hexdigest()
        if checksum != old_checksum:
            revision = f"{str(int(old_rev)+1)}-{checksum}"
            document[field] = revision

    @classmethod
    def to_db(cls, instance: BackendModel, json_dict: bool | None = True) -> dict:
        instance.updated_time = utcnow()
        return json.loads(instance.model_dump_json()) if json_dict else instance.model_dump()

    @classmethod
    def from_db(
        cls, model: Type[BackendModel], document: dict, json_dict: bool | None = True
    ) -> BackendModel:
        return (
            model.model_validate_json(json.dumps(document)) if json_dict else model.model_validate(document)
        )

    def get_uids(
        self,
        model: Type[BackendModel],
        skip: int = 0,
        limit: int = 25,
        query_filter: dict = None,
        sort: List = None,
        max_results: bool | None = False,
    ) -> Tuple[List[str], int] | List[str]:
        r = self.find(
            model=model,
            skip=skip,
            limit=limit,
            query_filter=query_filter,
            sort=sort,
            fields=["_id"],
            max_results=max_results,
            func=lambda x: x["_id"],
        )
        r: Tuple[List[str], int] | List[str]
        return r

    def get_instances(
        self,
        model: Type[BackendModel],
        skip: int = 0,
        limit: int = 0,
        query_filter: dict = None,
        sort: List = None,
        max_results: bool = False,
    ) -> Tuple[List[BackendModel], int] | List[BackendModel]:
        r = self.find(
            model=model,
            skip=skip,
            limit=limit,
            query_filter=query_filter,
            sort=sort,
            max_results=max_results,
            func=lambda x: self.from_db(model, x),
        )
        r: Tuple[List[BackendModel], int] | List[BackendModel]
        return r

    def indexes(
        self,
        model: Type[BackendModel],
        create_index_kwargs: dict | None,
        force_index_creation: bool = False,
    ) -> List[Index]:
        col_name = self.collection_name(model)
        if col_name not in self._indexes or force_index_creation:
            indexes = self.create_indexes(model, create_index_kwargs)
            self._indexes[col_name] = indexes
        return self._indexes[col_name]

    def get_index_by_name(self, model: Type[BackendModel], name: str) -> Index:
        col_name = self.collection_name(model)
        if col_name not in self._indexes:
            raise IndexNotExisting(model_name=col_name, name=name)
        model_indexes = self._indexes[col_name]
        index = next(filter(lambda x: x.name == name, model_indexes), None)
        if index is None:
            raise IndexNotExisting(model_name=col_name, name=name)
        return index

    def create_indexes(self, model: Type[BackendModel], create_index_kwargs: dict | None) -> List[Index]:
        if not hasattr(model, "Config"):
            return []

        if not hasattr(model.Config, "backend_indexes"):
            return []

        indexes = model.Config.backend_indexes
        for index in indexes:
            self.create_index(self.collection_name(model), index, **create_index_kwargs)
        return indexes

    def create_index(self, collection_name: str, index: Index, **kwargs):
        log.debug(f"[{collection_name}] Creating {index.type} index {index.name}...")

    @abstractmethod
    def delete_collection(self, model: Type[BackendModel]) -> None:
        # delete index info , for recreating it on next collection usage
        col_name = self.collection_name(model)
        if col_name in self._indexes:
            del self._indexes[col_name]

    @abstractmethod
    def post_instance(self, instance: BackendModel) -> BackendModel:  # pragma: nocover
        pass

    @abstractmethod
    def get_instance(self, model: Type[BackendModel], uid: str) -> BackendModel:  # pragma: nocover
        pass

    @abstractmethod
    def put_instance(self, instance: BackendModel, ignore_revision_conflict: bool = False) -> BackendModel:
        pass

    @abstractmethod
    def delete_uid(self, model: Type[BackendModel], uid: str):
        pass

    @abstractmethod
    def find(
        self,
        model: Type[BackendModel],
        skip: int = 0,
        limit: int = 25,
        query_filter: dict = None,
        sort: List = None,
        fields: List[str] = None,
        max_results: bool | None = False,
        func: Callable = None,
    ) -> Tuple[List[Any], int] | List[Any]:
        pass
